"""Database Module. Redshift and Athena functionalities"""

import json
import logging
import time
import urllib.parse
from typing import Dict, List, Optional, Tuple, Union
from urllib.parse import quote_plus

import boto3
import IPython.core.display
import pandas as pd
import pyathena
import sqlalchemy as sa
from IPython import get_ipython
from IPython.display import JSON
from pandas import DataFrame
from sqlalchemy.engine import create_engine
from sqlalchemy.orm import sessionmaker

from aws_orbit_sdk.common import *
from aws_orbit_sdk.glue_catalog import run_crawler
from aws_orbit_sdk.json import display_json
from aws_orbit_sdk.magics.database import AthenaMagics, RedshiftMagics

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger()

global __redshift__
global __athena__
__redshift__ = None
__athena__ = None


def check_in_jupyter():
    try:
        get_ipython()
        return True
    except NameError:
        return False


def get_redshift():
    global __redshift__
    if __redshift__ == None:
        __redshift__ = RedshiftUtils()
        if check_in_jupyter():
            ip = get_ipython()
            magics = RedshiftMagics(ip, __redshift__)
            ip.register_magics(magics)

    return __redshift__


def get_athena():
    global __athena__
    if __athena__ == None:
        __athena__ = AthenaUtils()
        if check_in_jupyter():
            ip = get_ipython()
            magics = AthenaMagics(ip, __athena__)
            ip.register_magics(magics)
    return __athena__


class DatabaseCommon:
    """
    Database functions, inherited by both Redshift and Athena classes.

    ...

    Attributes
    ----------
    current_engine : sqlalchemy.engine.Engine, optional

        A sql alchemy engine (default None).
    db_url : str, optional
        A sql alchemy connection string to a database (default None).
    redshift_role : str, optional
        The redshift role ARN that can be used to connect to Glue catalog (default None).
    db_class : str, optional
        The database class (default is None and is set when connecting to either Athena or Redshift).


    Methods
    -------
    def execute_ddl(self, ddl: str, namespace: Optional[Dict[str, str]] = dict()) -> None
        Executes a SQL ddl statement.

    def execute_query(self, sql: str, namespace: Optional[Dict[str, str]] = dict()) -> ResultProxy:
        Executes a SQL query.
    """

    current_engine = None
    db_url = None
    redshift_role = None
    db_class = None

    def execute_ddl(self, ddl: str, namespace: Optional[Dict[str, str]] = dict()) -> None:
        """
        Executes a SQL ddl statement.

        Parameters
        ----------
        ddl : str
            The data descrption language statement to execute in SQL.

        namespace : dict(), optional
            Mapping namespace keys to values if the values do not already exist.

        Returns
        -------
        None
            None.

        Examples
        --------
        >>> from aws.utils.notebooks.database import get_redshift
        >>> db_utils = get_redshift()
        >>> tableName = "myQuery1"
        >>> db_utils.execute_ddl(f'drop table if exists "mydatabase."{table_name}"')
        """

        with self.current_engine.connect() as conn:
            session = sessionmaker(bind=self.current_engine)()
            session.connection().connection.set_isolation_level(0)
            session.execute(ddl, namespace)
            conn.close()

    def execute_query(self, sql: str, namespace: Optional[Dict[str, str]] = dict()) -> sa.engine.result.ResultProxy:
        """
        Executes a SQL query.

        Parameters
        ----------
        sql : str
            SQL DML query.

        namespace : dict(), optional
            Mapping namespace keys to values if the values do not already exist.

        Returns
        -------
        rs : sqlalchemy.engine.result.ResultProxy
            Resulting data from sql query execution.

        Examples
        --------
        >>> from aws.utils.notebooks.database import get_redshift
        >>> db_utils = get_redshift()
        >>> db_utils.execute_query("select username from mydatabase.users")
        """

        with self.current_engine.connect() as conn:
            rs = self.current_engine.execute(sql, namespace)
            conn.close()
        return rs


class RedshiftUtils(DatabaseCommon):
    """
    Collection of Redshift functions used to easily connect and work with databases.

    Methods
    -------
    get_connection_to_redshift(self, clusterIdentifier, DbName, DbUser, lambdaName=None):
        Connect to an existing cluster or create a new cluster if it does not exists.

    create_external_schema(self, schema_name, glue_database)
         Creates external schema if it does not already exist from a glue data catalog database.

    create_external_table(self, select, database_name, table_name, format="Parquet", s3_location=None, options=""):
        Creates a new external table in the given database and runs a glue crawler to populate glue catalog tables with
        table metadata.

    connect_to_redshift(self, cluster_name, reuseCluster=True, startCluster=False, clusterArgs=dict()):
        Connects to a Redshift Cluster and returns connection information once redshift cluster is available for use.

    delete_redshift_cluster(self, cluster_name):
        Deletes a redshift cluster.

    get_redshift_functions(self):
        Provides a list of Redshift Functions that can be used.

    describe_redshift_function(self, funcName):
        Returns the environment variables for a specified redshift function.

    getCatalog(self, schema_name=None, table_name=None):
        Get Glue Catalog metadata of a specific Database table

    get_team_clusters(self, cluster_id=None):
        Retrieves Redshift Cluster information for the Team Cluster.

    """

    def get_connection_to_redshift(
        self,
        clusterIdentifier: str,
        DbName: str,
        DbUser: str,
        lambdaName: Optional[str] = None,
    ) -> Dict[str, Union[str, sa.engine.Engine]]:
        """
        Connect to an existing cluster or create a new cluster if it does not exists.

        Note
        ----
        To connect to existing cluster:
            **Cluster networking is configured correctly (has access to the Orbit Workbench VPC).
            **Cluster ESG was created, assigned to the cluster.
            **The ESG should have an ingress rule to allow inbound tcp on the redshift port from the Orbit Workbench
                instance SG.
            **A new teamspace should be launched from the Service Catalog and the ESG should be passed as a parameter
                to allow external traffic.

        Parameters
        ----------
        clusterIdentifier : str
            A unique name of the Redshift cluster, can be a new/existing cluster as well. If used with existing cluster,
            lambdaName should be given as well.
        DbName : str
            The Redshift DB name to connect to.
        DbUser : str
            The Redshift User name to connect with.
        lambdaName : str, optional
            For new cluster/mandatory for existing cluster connection, the lambda name which is responsible to get the
            cluster credentials.


        Returns
        -------
        db_url : str
            A sql alchemy connection string
        engine : sqlalchemy.engine.Engine
            A sql alchemy engine
        redshift_role : str
            The redshift role ARN that can be used to connect to Glue catalog

        (e.g. {'db_url': 'redshift+psycopg2://..', 'engine': engine,
        'redshift_role': 'arn:aws:iam::{accountid}:role/...'})

        Example
        --------
        >>> from aws.utils.notebooks.database import get_redshift
        >>> import json, urllib.parse
        >>> from sqlalchemy.engine import create_engine
        >>> response = get_redshift().get_connection_to_redshift(
        ...     clusterIdentifier = "my_cluster",
        ...     DbName = "my_database",
        ...     DBUser = "my_user",
        ...     lambdaName = None
        ...     )

        """

        redshift = boto3.client("redshift")

        if lambdaName:
            # Trying to get user and temp password from cluster
            try:
                lambda_client = boto3.client("lambda")
                response = lambda_client.invoke(
                    InvocationType="RequestResponse",
                    FunctionName=lambdaName,
                    Payload=json.dumps(
                        {
                            "DbUser": DbUser,
                            "DbName": DbName,
                            "clusterIdentifier": clusterIdentifier,
                        }
                    ),
                )

                # Parsing into JSON format
                data = json.loads(response["Payload"].read().decode("utf-8"))
                DbPassword = data["DbPassword"]
                DbUser = data["DbUser"]
            except Exception as e:
                logger.error(f"There was an error getting the cluster details: {data}")
                raise e
        else:
            response = redshift.get_cluster_credentials(
                DbUser=DbUser,
                DbName=DbName,
                AutoCreate=True,
                ClusterIdentifier=clusterIdentifier,
            )
            DbPassword = urllib.parse.quote(response["DbPassword"])
            DbUser = urllib.parse.quote(response["DbUser"])

        # Get the rest of the cluster properties
        clusterInfo = redshift.describe_clusters(ClusterIdentifier=clusterIdentifier)

        hostName = clusterInfo["Clusters"][0]["Endpoint"]["Address"]
        port = clusterInfo["Clusters"][0]["Endpoint"]["Port"]

        self.db_url = "redshift+psycopg2://{}:{}@{}:{}/{}".format(DbUser, DbPassword, hostName, port, DbName)
        if clusterInfo["Clusters"][0]["IamRoles"]:
            self.redshift_role = clusterInfo["Clusters"][0]["IamRoles"][0]["IamRoleArn"]
        else:
            self.redshift_role = None
        self.current_engine = create_engine(self.db_url)
        self.db_class = "redshift"
        return {
            "db_url": self.db_url,
            "engine": self.current_engine,
            "redshift_role": self.redshift_role,
        }

    def create_external_schema(self, schema_name: str, glue_database: str) -> None:
        """
        Creates external schema if it does not already exist from a glue data catalog database.

        Parameters
        ----------
        schema_name : str
            Name of the external schema that will be created.

        glue_database : str
            Name of the source database for creating the schema.

        Returns
        -------
        None
            None.

        Example
        -------
        >>> from aws.utils.notebooks.database import RedshiftUtils
        >>> RedshiftUtils.create_external_schema(schema_name="my_schema", glue_database="my_database")
        """

        createSchemaSql = f"create external schema IF NOT EXISTS {schema_name} from data catalog database '{glue_database}'\n iam_role '{self.redshift_role}'"
        self.execute_ddl(createSchemaSql, dict())

    def create_external_table(
        self,
        select: str,
        database_name: str,
        table_name: str,
        format: Optional[str] = "Parquet",
        s3_location: Optional[str] = None,
        options: Optional[str] = "",
    ) -> IPython.core.display.JSON:
        """
        Creates a new external table in the given database and runs a glue crawler to populate glue catalog tables with
        table metadata.

        Note
        ----
        This method function uses Amazon Redshift UNLOAD which splits the results of a select statement across a set of
        files, one or more files per node slice, to simplify parallel reloading of the data.

        Parameters
        ----------
        select : str
            SQL ddl statement for data to select (e.g. 'select * from table').
        database_name : str
            Name of database with existing data and schema.
        table_name : str
            The name of the table to be created.
        format : str, optional
            The file format for data files (default Parquet format).
        s3_location : str, optional
            The path to the Amazon S3 bucket or folder that contains the data files or a manifest file that contains a
            list of Amazon S3 object paths (used only if no database has no location).
        options : str, optional
            Specify additional table properties for when creating new table.

        Returns
        -------
        s : IPython.core.display.JSON
            An IPython JSON display representing the schema metadata for the table(s) / database

        Example
        --------
        >>> from aws.utils.notebooks.database import RedshiftUtils
        >>> from aws.utils.notebooks.json import display_json
        >>> RedshiftUtils.create_external_table(
        ...     select = 'select * from table1',
        ...     database_name = 'my_database',
        ...     table_name = 'myQuery1',
        ...     s3_location = 's3://bucketname/folder/'
        ... )
        """
        glue = boto3.client("glue")
        target_s3 = glue.get_database(Name=database_name)["Database"]["LocationUri"]
        s3 = boto3.client("s3")

        if target_s3 == None or len(target_s3) == 0:
            if s3_location != None:
                target_s3 = s3_location
            else:
                raise Exception("Database does not have a location , and location was not provided")
        bucket_name = target_s3[5:].split("/")[0]
        db_path = "/".join(target_s3[5:].split("/")[1:])
        s3_path = f"{target_s3}/{table_name}/"
        prefix = f"{db_path}/{table_name}/"
        response = s3.list_objects_v2(Bucket=bucket_name, Prefix=prefix)
        if "Contents" in response:
            for object in response["Contents"]:
                logger.info("Deleting {object['Key']}")
                s3.delete_object(Bucket=bucket_name, Key=object["Key"])

        ddl = f"""
            UNLOAD ('{select}')
                    TO '{s3_path}' iam_role '{self.redshift_role}'
                    FORMAT AS {format}
                    {options}
        """

        self.execute_ddl(ddl)

        logger.info("Query result s3 write complete.")

        run_crawler(f"crawler_for_{database_name}_{table_name}", database_name, s3_path)

        response = glue.get_table(DatabaseName=database_name, Name=table_name)

        recordCount = response["Table"]["Parameters"]["recordCount"]
        schema = response["Table"]["StorageDescriptor"]["Columns"]
        location = response["Table"]["StorageDescriptor"]["Location"]
        data_format = response["Table"]["Parameters"]["classification"]
        logger.info(f"Table {table_name} created: records: {recordCount}, format: {data_format}, location: {location}")
        s = {}
        for c in schema:
            s[c["Name"]] = c["Type"]
        return display_json(s, root="cols")

    def connect_to_redshift(
        self,
        cluster_name: str,
        reuseCluster: Optional[str] = True,
        startCluster: Optional[str] = False,
        clusterArgs: Optional[Dict[str, str]] = dict(),
    ) -> Dict[str, Union[str, sa.engine.Engine, bool]]:
        """
        Connects to a Redshift Cluster and returns connection information once redshift cluster is available for use.

        Parameters
        ----------
        cluster_name : str
            Name of the redshift cluster
        reuseCluster : bool, optional
            Boolean determining if you wish to reuse an existing cluster name (default True).
        startCluster : bool, optional
            Boolean determining if you wish to start a new cluster (default False)
        clusterArgs : dict(), optional
            Other redshift parameters you can optionally specify (e.g. auto_analyze, max_concurrency_scaling_clusters,
            statement_timeout, etc.)

        Returns
        -------
        db_url : str
            A sqlalchemy connection string.
        engine: sqlalchemy.engine.Engine
            A sql alchemy engine.
        cluster_identifier: str
            The unique identifier of the cluster.
        started: bool
            Boolean representing if cluster has started or not.
        redshift_role: str
            The redshift role ARN that can be used to access other AWS services when you execute the Amazon Redshift
            command.


        Examples
        --------
        >>> from aws.utils.notebooks.database import RedshiftUtils
        >>> from aws.utils.notebooks.common import get_properties
        >>> RedshiftUtils.connect_to_redshift(cluster_name= 'cluster-test')

        """

        props = get_properties()
        redshift = boto3.client("redshift")
        env = props["AWS_ORBIT_ENV"]
        team_space = props["AWS_ORBIT_TEAM_SPACE"]
        cluster_identifier = f"orbit-{env}-{team_space}-{cluster_name}".lower()
        try:
            clusters = redshift.describe_clusters(ClusterIdentifier=cluster_identifier)["Clusters"]
        except:
            clusters = []

        started = False
        if reuseCluster:
            if not clusters:
                if not startCluster:
                    raise Exception("cannot find running Redshift cluster: " + cluster_identifier)
                else:
                    (cluster_identifier, user) = self._start_and_wait_for_redshift(
                        redshift, cluster_identifier, props, clusterArgs
                    )
                    started = True
            else:
                logger.info("Using internal cluster")
                user = "master"
        else:
            if not startCluster:
                raise Exception("If reuseCluster is False , then startCluster must be True")
            else:
                (cluster_identifier, user) = self._start_and_wait_for_redshift(
                    redshift, cluster_identifier, props, clusterArgs
                )
                started = True

        conn = self.get_connection_to_redshift(cluster_identifier, "defaultdb", user)

        return {
            "db_url": conn["db_url"],
            "engine": conn["engine"],
            "cluster_identifier": cluster_identifier,
            "started": started,
            # 'user': user,
            # 'password': password,
            "redshift_role": conn["redshift_role"],
        }

    def _get_cred_to_redshift_cluster(self, cluster_name: str) -> Dict[str, str]:

        """
        Invokes Lambda Function to Connect to a Redshift Cluster and returns user credentials (username and password)
        upon successful connection.

        """
        props = get_properties()
        funcName = "ConnectToRedshiftFunction"

        orbit = props["AWS_ORBIT_ENV"]
        team_space = props["AWS_ORBIT_TEAM_SPACE"]
        functionName = "{}-{}-{}".format(orbit, team_space, funcName)
        lambda_client = boto3.client("lambda")

        invoke_response = lambda_client.invoke(
            FunctionName=functionName,
            Payload=bytes(json.dumps({"cluster_name": cluster_name}), "utf-8"),
            InvocationType="RequestResponse",
            LogType="Tail",
        )
        response_payload = json.loads(invoke_response["Payload"].read().decode("utf-8"))
        if "200" != response_payload["statusCode"]:
            logger.error(response_payload)
            raise Exception("could not connect to cluster")

        return {
            "user": response_payload["user"],
            "password": response_payload["password"],
        }

    def delete_redshift_cluster(self, cluster_name: str) -> None:

        """
        Deletes a redshift cluster.

        Parameters
        ----------
        cluster_name : str
            The Redshift cluster name given when creating the cluster

        Returns
        -------
        None
            None.

        Example
        --------
        >>> from aws.utils.notebooks.database import RedshiftUtils
        >>> from aws.utils.notebooks.common import get_properties
        >>> RedshiftUtils.delete_redshift_cluster(cluster_name = "my_cluster")

        """

        props = get_properties()
        env = props["AWS_ORBIT_ENV"]
        team_space = props["AWS_ORBIT_TEAM_SPACE"]
        redshift = boto3.client("redshift")
        namespace = f"orbit-{env}-{team_space}-"
        cluster_name_value = cluster_name.lower()
        cluster_identifier = cluster_name if namespace in cluster_name_value else namespace + cluster_name_value
        res = redshift.delete_cluster(ClusterIdentifier=cluster_identifier, SkipFinalClusterSnapshot=True)

        if "errorMessage" in res:
            logger.error(res["errorMessage"])
        else:
            logger.info("Cluster termination started")

        return

    def get_redshift_functions(self) -> List[str]:
        """
        Provides a list of Redshift Functions that can be used.

        Parameters
        ----------
        None
            None.

        Returns
        -------
        functions: list
            List of Redshift Functions

        Examples
        --------
        >>> from aws.utils.notebooks.database import get_redshift
        >>> get_redshift().get_redshift_functions()
        """

        return list(self._get_redshift_functions().keys())

    def describe_redshift_function(self, funcName: str) -> Dict[str, str]:
        """
        Returns the environment variables for a specified redshift function.

        Parameters
        ----------
        funcName : str
            Name of a redshift function

        Returns
        -------
        description : Dict[str, str]
            Key-Value pairs of names and descriptions of the given redshift function environment variables.

        Examples
        --------
        >>> from aws.utils.notebooks.database import get_redshift
        >>> get_redshift().describe_redshift_function(funcName = 'get_connection_to_redshift')
        """

        return self._get_redshift_functions()[funcName]

    def _get_redshift_functions(self) -> Dict[str, Dict[str, str]]:
        """
        Returns a dictionary of all redshift function names  with their parameters/configuration
        using Lambda.list_functions().
        """

        lambda_client = boto3.client("lambda")
        props = get_properties()
        env_name = props["AWS_ORBIT_ENV"]
        team_space = props["AWS_ORBIT_TEAM_SPACE"]
        namespace = f"{env_name}-{team_space}"
        funcs = []
        while True:
            response = lambda_client.list_functions()
            token = response["NextMarker"] if "NextMarker" in response else None
            for func in response["Functions"]:
                if (
                    namespace in func["FunctionName"]
                    and "Environment" in func
                    and "Variables" in func["Environment"]
                    and "RedshiftClusterParameterGroup" in func["Environment"]["Variables"]
                    and "StartRedshift" in func["FunctionName"]
                ):
                    funcs.append(func)
            if token == None:
                break

        func_descs = {}

        for f in funcs:
            user_name = f["FunctionName"][f["FunctionName"].index("StartRedshift") + len("StartRedshift-") :]
            func_desc = f["Environment"]["Variables"]
            del func_desc["RedshiftClusterParameterGroup"]
            del func_desc["RedshiftClusterSubnetGroup"]
            del func_desc["RedshiftClusterSecurityGroup"]
            del func_desc[ORBIT_ENV]
            del func_desc[AWS_ORBIT_TEAM_SPACE]
            del func_desc["SecretId"]
            del func_desc["PortNumber"]
            del func_desc["Role"]
            func_descs[user_name] = func_desc

        return func_descs

    def _start_and_wait_for_redshift(
        self,
        redshift: boto3.client("redshift"),
        cluster_name: str,
        props: Dict[str, str],
        clusterArgs: Dict[str, str],
    ) -> Tuple[str, str]:
        """
        Starts the Redshift Cluster and waits until cluster is available for use.

        """
        env = props["AWS_ORBIT_ENV"]
        team_space = props["AWS_ORBIT_TEAM_SPACE"]
        funcName = f"Standard"
        if "redshift_start_function" in clusterArgs.keys():
            funcName = clusterArgs["redshift_start_function"]

        cluster_def_func = f"orbit-{env}-{team_space}-StartRedshift-{funcName}"

        lambda_client = boto3.client("lambda")
        clusterArgs["cluster_name"] = cluster_name
        invoke_response = lambda_client.invoke(
            FunctionName=cluster_def_func,
            Payload=bytes(json.dumps(clusterArgs), "utf-8"),
            InvocationType="RequestResponse",
            LogType="Tail",
        )

        response_payload = json.loads(invoke_response["Payload"].read().decode("utf-8"))
        if "statusCode" not in response_payload or "200" != response_payload["statusCode"]:
            logger.error(response_payload)
            raise Exception("could not start cluster")

        cluster_id = response_payload["cluster_id"]
        user = response_payload["username"]
        logger.info("cluster created: %s", cluster_id)
        logger.info("waiting for created cluster: %s", cluster_id)
        waiter = redshift.get_waiter("cluster_available")
        waiter.wait(ClusterIdentifier=cluster_id, WaiterConfig={"Delay": 60, "MaxAttempts": 15})
        time.sleep(60)  # allow DB to for another min
        return (cluster_id, user)

    def create_cluster(self, cluster_name: str, number_of_nodes: str, node_type: str) -> Dict[str, str]:
        """
        Creates a Redshift Cluster.

        """
        props = get_properties()
        env = props["AWS_ORBIT_ENV"]
        team_space = props["AWS_ORBIT_TEAM_SPACE"]
        cluster_def_func = f"orbit-{env}-{team_space}-StartRedshift-Standard"
        cluster_identifier = f"orbit-{env}-{team_space}-{cluster_name}".lower()

        cluster_args = {"cluster_name": cluster_identifier, "Nodes": number_of_nodes, "NodeType": node_type}

        lambda_client = boto3.client("lambda")
        invoke_response = lambda_client.invoke(
            FunctionName=cluster_def_func,
            Payload=bytes(json.dumps(cluster_args), "utf-8"),
            InvocationType="RequestResponse",
            LogType="Tail",
        )

        response_payload = json.loads(invoke_response["Payload"].read().decode("utf-8"))
        if "statusCode" not in response_payload or "200" != response_payload["statusCode"]:
            logger.error(response_payload)
            error_message = response_payload["errorMessage"]
            response = {"status": "400", "message": f"Error creating Redshift cluster - {error_message}"}
        else:
            cluster_id = response_payload["cluster_id"]
            logger.info("cluster created: %s", cluster_id)
            response = {
                "status": str(response_payload["statusCode"]),
                "message": f"Successfully created Redshift cluster {cluster_id}",
            }

        return response

    def _add_json_schema(
        self,
        s3: boto3.client("s3"),
        glue: boto3.client("glue"),
        table: str,
        database_name: str,
        table_name: str,
    ) -> None:
        """
        Updates schema of specified table based on Glue Catalog metadata.

        """

        response = glue.get_table(DatabaseName=database_name, Name=table_name)
        # This is another way to get the properties. however, for now we will stick with the glue way that can also work in the future for Athena
        # """
        # SELECT  tablename, parameters
        #         FROM (
        #         SELECT btrim(ext_tables.tablename::text)::character varying(128) AS tablename,
        #         btrim(ext_tables.parameters::text)::character varying(500) AS parameters
        #         FROM pg_get_external_tables() ext_tables(esoid integer, schemaname character varying, tablename character varying, "location" character varying, input_format character varying,
        #         output_format character varying, serialization_lib character varying, serde_parameters character varying, compressed integer, parameters character varying)
        #         ) A
        #         LIMIT 10
        #         ;
        #
        # """

        if "Parameters" in response["Table"]:
            params = response["Table"]["Parameters"]
            if "json.schema" in params:
                schema_path = params["json.schema"]
                (bucket, key) = split_s3_path(schema_path)
                try:
                    s3_obj = s3.get_object(Bucket=bucket, Key=key)
                    content = s3_obj["Body"].read().decode("utf-8")
                    schema = json.loads(content)
                    table["jschema"] = schema
                except Exception as e:
                    logger.error(str(e))
                    raise Exception(f"Cannot access {table_name} schema file at: {schema_path}")

    def getCatalog(
        self, schema_name: Optional[str] = None, table_name: Optional[str] = None
    ) -> IPython.core.display.JSON:
        """
        Get Glue Catalog metadata of a specific Database table.

        Parameters
        ----------
        schema_name : str, optional
            Name of schema to retrieve Glue Catalog for (default looks at all schema with give tablenames in the
            current database)

        table_name : str, optional
            Name of table to retrieve Glue Catalog for (default looks at all tables with give schemanames in the
            current database)


        Returns
        -------
        schema : IPython.core.display.JSON
            An IPython JSON display representing the schema metadata for the table(s) / database

        Example
        --------
        >>> from aws.utils.notebooks.database import RedshiftUtils
        >>> from aws.utils.notebooks.json import display_json
        >>> RedshiftUtils.getCatalog(schema_name="my_schema",table_name="table1")
        """

        glue = boto3.client("glue")
        s3 = boto3.client("s3")
        sql = """
                    SELECT cols.schemaname, cols.tablename, cols.columnname, cols.external_type, cols.columnnum, tables.location, schemas.databasename, tables.parameters
                     FROM SVV_EXTERNAL_COLUMNS cols natural join SVV_EXTERNAL_TABLES tables natural join SVV_EXTERNAL_SCHEMAS schemas
                """

        if schema_name or table_name:
            sql += "WHERE "
            if schema_name and table_name:
                sql += f"schemaname = '{schema_name}' AND tablename = '{table_name}'"
            elif schema_name:
                sql += f"schemaname = '{schema_name}'"
            elif table_name:
                sql += f"tablename = '{table_name}'"

        sql += "\n order by schemas.schemaname, tables.tablename, columnnum\n"

        rs = self.current_engine.execute(sql)
        res = rs.fetchall()
        if len(res) == 0:
            print("no external schema or tables found")
            return

        df = DataFrame(res)
        df.columns = rs.keys()
        schemas = dict()
        for row in df.itertuples():
            if row[1] not in schemas:
                schemas[row[1]] = dict()
            if row[2] not in schemas[row[1]]:
                schemas[row[1]][row[2]] = dict()
                table = schemas[row[1]][row[2]]
                table["name"] = row[2]
                table["location"] = row[6]
                table["cols"] = dict()
                self._add_json_schema(s3, glue, table, row[7], row[2])
            table = schemas[row[1]][row[2]]
            colInfo = dict()
            colInfo["name"] = row[3]
            colInfo["type"] = row[4]
            colInfo["order"] = row[5]
            table["cols"][row[3]] = colInfo

        return display_json(schemas, root="glue databases")

    def get_team_clusters(
        self, cluster_id: Optional[str] = None
    ) -> Dict[str, Dict[str, Union[str, Dict[str, Union[str, int]]]]]:
        """
        Retrieves Redshift Cluster information for the Team Cluster.

        Parameters
        ----------
        cluster_id : str, optional
            Gets information for a specific cluster Id. Default looks at cluster tagged with 'AWS_ORBIT_TEAM_SPACE'

        Returns
        -------
        clusters_info : Dict[str, Dict[str, Union[str, Dict[str, Union[str, int]]]]]
            Information on the cluster(s) and their configuration.

        Example
        --------
        >>> from aws.utils.notebooks.database import RedshiftUtils
        >>> from aws.utils.notebooks.common import get_workspace
        >>> RedshiftUtils.get_team_clusters(cluster_id= "my_cluster")

        """

        redshift = boto3.client("redshift")
        props = get_properties()
        if cluster_id == None:
            clusters = redshift.describe_clusters(TagValues=[props["AWS_ORBIT_TEAM_SPACE"]])["Clusters"]
        else:
            clusters = redshift.describe_clusters(
                ClusterIdentifier=cluster_id,
            )["Clusters"]
        clusters_info = {}
        for cluster in clusters:
            cluster_id = cluster["ClusterIdentifier"]
            cluster_model = {}
            cluster_model["cluster_id"] = cluster_id
            cluster_model["Name"] = cluster_id
            cluster_model["State"] = cluster["ClusterStatus"]
            if "Endpoint" in cluster and "Address" in cluster["Endpoint"]:
                cluster_model["ip"] = f"{cluster['Endpoint']['Address']}:{cluster['Endpoint']['Port']}"

            cluster_nodes_info = {
                "node_type": cluster["NodeType"],
                "nodes": len(cluster["ClusterNodes"]),
            }
            cluster_model["instances"] = cluster_nodes_info
            clusters_info[cluster_id] = cluster_model
            cluster_model["info"] = cluster
        return clusters_info


class AthenaUtils(DatabaseCommon):
    """Collection of Athena functions used to easily connect and work with databases.

    Methods
    -------
    get_connection_to_athena(self, DbName, region_name=None, S3QueryResultsLocation=None):
        Connect Athena to an existing database.

    getCatalog(self, database=None):
        Get Data Catalog of a specific Database


    """

    def get_connection_to_athena(
        self,
        DbName: str,
        region_name: Optional[str] = None,
        S3QueryResultsLocation: Optional[str] = None,
    ) -> Dict[str, Union[str, sa.engine.Engine]]:
        """
        Connect Athena to an existing database

        Parameters
        ----------
        DbName : str
            Name of the glue database name.

        region_name : str, optional
            The region to connect to athena. The default region will be used if receives None.

        S3QueryResultsLocation : str, optional
            The s3 bucket where to store query results. The results will not be saved if received None.


        Returns
        -------
        db_url : str
            A sql alchemy connection string.
        engine : sqlalchemy.engine.Engine
            A sql alchemy engine.

        Example
        --------
        >>> from aws.utils.notebooks.database import AthenaUtils
        >>> from sqlalchemy.engine import create_engine
        >>> from aws.utils.notebooks.common import get_workspace
        >>> (db_url,engine) = AthenaUtils.get_connection_to_athena(
        ...     DbName = glue_db,
        ...     my_region = my_region,
        ...     S3QueryResultsLocation = results_location)
        """

        workspace = get_workspace()
        if region_name == None:
            region_name = workspace["region"]

        if S3QueryResultsLocation == None:
            S3QueryResultsLocation = f"{workspace['ScratchBucket']}/athena"

        template_con_str = (
            "awsathena+rest://athena.{region_name}.amazonaws.com:443/" "{schema_name}?s3_staging_dir={s3_staging_dir}"
        )
        conn_str = template_con_str.format(
            region_name=region_name,
            schema_name=DbName,
            s3_staging_dir=quote_plus(S3QueryResultsLocation),
        )

        engine = create_engine(conn_str)
        self.db_url = conn_str
        self.current_engine = engine
        self.db_class = "athena"
        return {
            "db_url": self.db_url,
            "engine": self.current_engine,
        }

    def getCatalog(self, database: Optional[str] = None) -> IPython.core.display.JSON:
        """
        Get Data Catalog of a specific Database

        Parameters
        ----------
        database : str
            Name of database to catalog.

        Returns
        -------
         schema : IPython.core.display.JSON
            An IPython JSON display representing the schema metadata for a database

        Example
        --------
        >>> from aws.utils.notebooks.database import AthenaUtils
        >>> from aws.utils.notebooks.json import display_json
        >>> AthenaUtils.getCatalog(database="my_database")
        """

        glue = boto3.client("glue")
        schemas = dict()
        response = glue.get_tables(DatabaseName=database, MaxResults=1000)
        for t in response["TableList"]:
            table = dict()
            schemas[t["Name"]] = table
            table["name"] = t["Name"]
            table["location"] = t["StorageDescriptor"]["Location"]
            table["cols"] = dict()
            for c in t["StorageDescriptor"]["Columns"]:
                col = dict()
                table["cols"][c["Name"]] = col
                col["name"] = c["Name"]
                col["type"] = c["Type"]

        return display_json(schemas, root="glue databases")

    def get_sample_data(self, database: str, table: str, sample: int, field: str, direction: str):
        workspace = get_workspace()
        logger.info(f"query staging location: {workspace['ScratchBucket']}/athena/query/")
        conn = pyathena.connect(
            s3_staging_dir=f"{workspace['ScratchBucket']}/athena/query/",
            region_name=workspace["region"],
        )
        if field and len(field) > 0:
            query = f'SELECT * FROM "{database}"."{table}" order by {field} desc LIMIT {sample}'
        else:
            query = f'SELECT * FROM "{database}"."{table}" LIMIT {sample}'
        df = pd.read_sql(query, conn)
        result = df.to_json(orient="records")
        return result



import argparse
import json
import sys
from typing import Any, Dict, List, Optional, Union

import IPython.core.display
from IPython import get_ipython
from IPython.core.magic import Magics, cell_magic, line_cell_magic, line_magic, magics_class, needs_local_scope

from aws_orbit_sdk import controller
from aws_orbit_sdk.json import display_json, run_schema_induction_args


def exception_handler(exception_type, exception, traceback):
    print("%s: %s" % (exception_type.__name__, exception), file=sys.stderr)


class ArgumentParserNoSysExit(argparse.ArgumentParser):
    def error(self, message):
        self.print_help(sys.stderr)
        raise Exception(message)


@magics_class
class OrbitWorkbenchMagics(Magics):
    def __init__(self, shell):
        super(OrbitWorkbenchMagics, self).__init__(shell)
        self.ip = get_ipython()

    @line_magic
    def schema_induction(self, line: str) -> Dict[str, Union[str, Dict[str, str]]]:
        """
        Calls on run_process to run Schema Induction with given user arguments gets ddl and schema metadata for a
        specified table.

        Parameters
        ----------
        User parameters including:
        '-i' : str
            An input json file path.
        '-c' : str, optional
            Specify compute service to run schema induction (default ec2).
        '-t' : str
            Table name to use when creating DDL.
        '--location' : str
            Table s3 location to use when creating DDL.
        '--root' : str
            The root directory name of the data.
        '-a' : bool, optional
            Is the document a json array (default True).

        Returns
        ----------
        ddl : str
            SQL ddl statement to create new external table with given metadata.
        schema :
            Schema metadata stored as a json for the specified table.

        Example
        ----------
        >>> %schema_induction -i data_path -c ec2 -t table_name --location s3_location --root ClaimData
        """

        lineArgs = line.split(" ")
        return run_schema_induction_args(lineArgs)

    # @needs_local_scope
    # @line_magic
    # def display_grid(self, line: str, local_ns: Optional[Dict[str, str]] = None) -> qgrid.QGridWidget:
    #     """
    #     Renders a DataFrame or Series as an interactive qgrid, represented by an instance of the QgridWidget class.
    #
    #     Parameter
    #     ----------
    #     line : str
    #         Specifies the name of the DataFrame
    #     local_ns : str
    #         The namespace to load into IPython user namespace
    #
    #     Return
    #     ----------
    #     qgrid : qgrid.QGridWidget
    #         A DataFrame or Series shown as an interactive qgrid
    #
    #     Example
    #     ----------
    #     >>> %%sql analysis <<
    #     >>> select * from database.table1
    #
    #     >>> %display_grid analysis
    #     """
    #     lineArgs = line.split(" ")
    #     dataset = local_ns[lineArgs[0]]
    #     return qgrid.show_grid(
    #         dataset.DataFrame(),
    #         show_toolbar=False,
    #         grid_options={
    #             "forceFitColumns": False,
    #             "defaultColumnWidth": 140,
    #             "editable": False,
    #             "autoEdit": False,
    #         },
    #     )

    @needs_local_scope
    @line_magic
    def display_tree(self, line: str, local_ns: Optional[Dict[str, str]] = None) -> IPython.core.display.JSON:
        """

        Parameters
        ----------
        line : str
            Dataset name with data to display
        local_ns : str
            The namespace to load into IPython user namespace

        Returns
        ----------
        tree: IPython.core.display.JSON
        An IPython JSON display representing the data.

        Example
        ----------
        >>> %display_tree dataset
        """

        lineArgs = line.split(" ")
        dataset = local_ns[lineArgs[0]]
        return display_json(dataset)

    @line_magic
    def short_errors(self, line):
        ip = get_ipython()
        ip._showtraceback = exception_handler

    @cell_magic
    def schedule_notebook(self, line: str, cell: str, local_ns: Optional[Dict[str, str]] = None) -> str:
        """
        Schedule a notebook execution.

        Parameters
        ----------
        Line parameters:
            '-cron' : str
                specify cron-based schedule
            '-id' : str
                specify unique identifier for this scheduled task

        Cell parameters:
            'taskConfiguration' : dict
                A task definition to execute
                'notebooks' : list
                    A list of notebook task definition to run
                    'notebookName' : str
                        The filename of the notebook
                    'sourcePath' : str
                        The relative path to the notebook file starting at the repository root
                    'targetPath' : str
                        The target S3 directory where the output notebook and all related output will be generated
                    'params' : dict
                        A list of parameters for this task to override the notebook parameters
                'compute' : dict, optional
                    A list of runtime parameters to control execution
                    'container' : dict
                        A list of parameters to control container execution
                        'p_concurrent' : str
                            The number of parallel threads inside the container that will execute notebooks
                'env_vars' : list, optional
                    A list of environment parameters to pass to the container
                    'container' : dict
                        A list of parameters to control container execution
                        'p_concurrent' : str
                            The number of parallel threads inside the container that will execute notebooks
        local_ns : str
            The namespace to load into IPython user namespace

        Returns
        --------
        response : str
            The response will be ARN of the event rule created to start this execution.

        Example
        -------
        >>> %%schedule_notebook -cron cron(0/3 * 1/1 * ? *)  -id yyy
        >>> {
        >>>      "tasks":  [
        >>>            {
        >>>                  "notebookName": "Example-1-SQL-Analysis-Athena.ipynb",
        >>>                  "sourcePath": "samples/notebooks/B-DataAnalyst",
        >>>                  "targetPath": "tests/Z-Tests/Scheduling",
        >>>                  "targetPrefix": "yyy",
        >>>                  "params": {
        >>>                        "glue_db" : "cms_raw_db",
        >>>                        "target_db" : "users"
        >>>                  }
        >>>            }
        >>>      ]
        >>> }
        """
        if len(line) == 0:
            print("must provide -cron and -id parameters")
            return

        parser = ArgumentParserNoSysExit(description="schedule a task to be executed on container")
        parser.add_argument("-cron", required=True, nargs="+", help="specify cron-based schedule")

        parser.add_argument(
            "-id",
            required=True,
            help="specify unique identifier for this scheduled task",
        )

        try:
            args = parser.parse_args(line.strip().split(" "))
            cronStr = " ".join(args.cron)
            return controller.schedule_notebooks(
                triggerName=args.id,
                frequency=cronStr,
                taskConfiguration=json.loads(cell),
            )
        except Exception as e:
            print("Error!")
            print(str(e))

    @cell_magic
    def run_notebook(self, line: Optional[str], cell: str, local_ns: Optional[Dict[str, str]] = None) -> List[str]:
        """
        Run a notebook execution.

        Parameters
        ----------
        local_ns : str
            The namespace to load into IPython user namespace

        Cell parameters:
            'taskConfiguration' : dict
                A task definition to execute
                'notebooks' : list
                    A list of notebook task definition to run
                    'notebookName' : str
                        The filename of the notebook
                    'sourcePath' : str
                        The relative path to the notebook file starting at the repository root
                    'targetPath' : str
                        The target S3 directory where the output notebook and all related output will be generated
                    'params' : dict
                        A list of parameters for this task to override the notebook parameters
                'compute' : dict, optional
                    A list of runtime parameters to control execution
                    'container' : dict
                        A list of parameters to control container execution
                        'p_concurrent' : str
                            The number of parallel threads inside the container that will execute notebooks
                    'sns.topic.name' : str
                        A name of a topic to which messages are sent on task completion or failure
                'env_vars' : list, optional
                    A list of environment parameters to pass to the container

        Returns
        --------
        response: list
            List of containers ARNs that are started
            (e.g. ['arn:aws:ecs:us-east-1:{accountid}:task/a99f186c-160d-48b8-99de-a51f40f7c78e'])

        Example
        -------
        >>> %%run_notebook
        >>> taskConfiguration = {
        >>> "notebooks":  [ {
        >>>     "notebookName": "Example-2-Extract-Files.ipynb",
        >>>     "sourcePath": "samples/notebooks/A-LakeCreator",
        >>>     "targetPath": "tests/createLake",
        >>>     "params": {
        >>>         "bucketName": bucketName,
        >>>         "zipFileName": file,
        >>>         "targetFolder": extractedFolder
        >>>         },
          ...
        >>> }],
        >>> "compute": {
        >>>     "container" : {
        >>>         "p_concurrent": "4",
        >>>         },
        >>>     "env_vars": [
        >>>                 {
        >>>                     'name': 'cluster_name',
        >>>                     'value': clusterName
        >>>                 }
        >>>     ],
        >>>     "sns.topic.name": 'TestTopic',
        >>> }}
        """
        try:
            return controller.run_notebooks(taskConfiguration=json.loads(cell))
        except Exception as e:
            print("Error!")
            print(str(e))

    @cell_magic
    def run_python(self, line: Optional[str], cell: str, local_ns: Optional[Dict[str, str]] = None) -> List[str]:
        """
        Run some python code.

        Parameters
        ----------
        local_ns : str
            The namespace to load into IPython user namespace

        Cell parameters:
            'taskConfiguration' : dict
                    A task definition to execute
                    'tasks' : list
                        A list of python task definition to run
                        'module' : str
                            The python module to run (without .py ext)
                        'functionName' : str
                            The python function to start the execution
                        'sourcePaths' : list
                            A list of s3 python source paths used for importing packages or modules into the application
                        'params' : dict
                            A list of parameters for this task to override the notebook parameters
                'compute' : dict, optional
                    A list of runtime parameters to control execution
                    'container' : dict
                        A list of parameters to control container execution
                        'p_concurrent' : str
                            The number of parallel threads inside the container that will execute notebooks
                'env_vars' : list, optional
                    A list of environment parameters to pass to the container
                    'container' : dict
                        A list of parameters to control container execution
                        'p_concurrent' : str
                            The number of parallel threads inside the container that will execute notebooks

        Returns
        -------
        response : list
            A list of containers ARNs that are started
            (e.g. ['arn:aws:ecs:us-east-1:{accountid}:task/a9a186c-160d-48b8-99de-a51f40f7c78e'])

        Example
        -------
        >>> taskConfiguration = {
        >>> "tasks":  [
        >>>     {
        >>>       "module": "pyspark.run_pyspark_local",
        >>>       "functionName": "run_spark_job",
        >>>       "sourcePaths": ["DataScienceRepo/samples/python"],
        >>>       "params": {
        >>>                 "bucket": "users-env2",
        >>>                 "p2": 'bar'
        >>>       }
        >>>     }
        >>> ],
        >>> "compute": {
        >>>     "container" : {
        >>>         "p_concurrent": "4"
        >>> },
        >>> "env_vars": [
        >>>             {
        >>>                 'name': 'cluster_name',
        >>>                 'value': clusterName
        >>>             }
        >>> ]
        >>> }}
        """
        try:
            return controller.run_python(taskConfiguration=json.loads(cell))
        except Exception as e:
            print("Error!")
            print(str(e))

    @line_magic
    def delete_schedule_task(self, line: str) -> None:
        """
        Deletes a scheduled task execution.

        Parameters
        ----------
        '-id' : str
            The arn of the event rule

        Returns
        -------
        None
            None.

        Example
        -------
        >>> %delete_schedule_task -id taskid
        """
        if len(line) == 0:
            print("must provide -id parameter")
            return
        parser = ArgumentParserNoSysExit(description="delete schedule task")
        parser.add_argument(
            "-id",
            required=True,
            help="specify unique identifier for this scheduled task",
        )
        try:
            args = parser.parse_args(line.strip().split(" "))
            return controller.delete_task_schedule(triggerName=args.id)
        except Exception as e:
            print("Error!")
            print(str(e))


# In order to actually use these magics, you must register them with a
# running IPython.
def load_ipython_extension(ipython):
    """
    Any module file that define a function named `load_ipython_extension`
    can be loaded via `%load_ext module.path` or be configured to be
    autoloaded by IPython at startup time.
    """
    # You can register the class itself without instantiating it.  IPython will
    # call the default constructor on it.
    ipython.register_magics(OrbitWorkbenchMagics)


#
# def apt_completers(self, event):
#     """ This should return a list of strings with possible completions.
#
#     Note that all the included strings that don't start with event.symbol
#     are removed, in order to not confuse readline.
#     """
#     return ['update', 'upgrade', 'install', 'remove']
#
#     completerlib.quick_completer('SELECT', ['FROM','WHERE', 'GROUPBY'])
#
#


ip = get_ipython()
magics = OrbitWorkbenchMagics(ip)
ip.register_magics(magics)



import setuptools

with open("VERSION", "r") as version_file:
    version = version_file.read().strip()

with open("README.md") as fp:
    long_description = fp.read()


setuptools.setup(
    name="aws-orbit-sdk",
    version=version,
    author="AWS Professional Services",
    author_email="aws-proserve-opensource@amazon.com",
    url="https://github.com/awslabs/aws-orbit-workbench",
    project_urls={"Org Site": "https://aws.amazon.com/professional-services/"},
    description="AWS Orbit Workbench SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["aws", "cdk"],
    package_dir={"aws_orbit_sdk": "aws_orbit_sdk"},
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=[
        "boto3>=1.12.49,<2.0.0",
        "pyyaml~=5.4",
        "ipython>=7.16.0,<7.20.0",
        "pandas>=1.1.0,<=1.2.0",
        "psycopg2-binary~=2.8.4",
        "SQLAlchemy>=1.3.10,<1.3.16",
        "sqlalchemy-redshift~=0.7.5",
        "requests>=2.24.0,<2.26.0",
        "kubernetes~=12.0.1",
        "python-slugify~=4.0.1",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: JavaScript",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
)



import json
import logging
import os
import socket
import time
from typing import Any, Dict, List, Mapping, Optional, Tuple

import boto3
import requests

from aws_orbit_sdk.common import *

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger()

SSM_PARAMETER_PREFIX = "/emr_launch/emr_launch_functions"


def get_virtual_cluster_id() -> str:
    emr = boto3.client("emr-containers")
    props = get_properties()
    env_name = props["AWS_ORBIT_ENV"]
    team_space = props["AWS_ORBIT_TEAM_SPACE"]
    response = emr.list_virtual_clusters(
        containerProviderId=f"orbit-{env_name}", containerProviderType="EKS", states=["RUNNING"], maxResults=500
    )
    if "virtualClusters" not in response or len(response["virtualClusters"]) == 0:
        raise Exception("Virtual EMR Cluster not found")
    for c in response["virtualClusters"]:
        if c["name"] == f"orbit-{env_name}-{team_space}":
            return c["id"]


def stop_cluster(cluster_id: str) -> None:
    """
    Stops a running EMR cluster.

    Parameters
    ----------
    cluster_id : str
        The EMR cluster ID to terminate

    Returns
    -------
    None
        None.

    Example
    -------
    >>> import aws.utils.notebooks.spark.emr as sparkConnection
    >>> sparkConnection.stop_cluster(cluster_id=cluster_id)
    """

    emr = boto3.client("emr")

    response = emr.terminate_job_flows(JobFlowIds=[cluster_id])
    logger.info("stopping cluster %s", cluster_id)
    logger.info(response)
    return None


def connect_to_spark(
    cluster_name: Optional[str] = None,
    reuseCluster: Optional[bool] = True,
    startCluster: Optional[bool] = False,
    clusterArgs: Optional[Dict[str, str]] = dict(),
    waitWhenResizing: Optional[bool] = True,
) -> Tuple[str, str, bool]:
    """
    Returns a livy URL that can be used to start a new spark session. The API can also create a new cluster if it does
    not exists.

    Parameters
    ----------
    cluster_name : str, optional
        A unique name for the EMR cluster
    reuseCluster : string, optional
        If reuseCluster is True and cluster exists , will reuse the existing cluster (default: True)
    startCluster : string, optional
        If startCluster is True and cannot re-use cluster, then will start a new cluster (default: False)
    clusterArgs : dict, optional
        Optional list of arguments to control the EMR definition used to start the new cluster
    waitWhenResizing : bool, optional
        If waitWhenResizing is True and the cluster task nodes are resizing, the call will wait until operation is completed

    Returns
    -------
    livy_url : str
        The livy url to use for starting the spark session
    cluster_id : str
        The EMR cluster ID used in the session
    started : bool
        True if a new cluster was started in this call

    Example
    -------
    >>> import aws.utils.notebooks.spark.emr as sparkConnection
    >>> import sparkmagic.utils.configuration as conf
    >>> conf.override(conf.livy_session_startup_timeout_seconds.__name__, 240)
    >>> (livy_url, cluster_id, started) = sparkConnection.connect_to_spark(clusterName=clusterName,
    ...                                                            reuseCluster=True,
    ...                                                            startCluster=True,
    ...                                                            clusterArgs={},
    ...                                                            waitWhenResizing=True)
    """
    props = get_properties()
    if cluster_name == None:
        if "ClusterName" in clusterArgs:
            cluster_name = clusterArgs["ClusterName"]

    emr = boto3.client("emr")
    ClusterStates = ["STARTING", "BOOTSTRAPPING", "RUNNING", "WAITING"]
    res = emr.list_clusters(ClusterStates=ClusterStates)
    while True:
        clusters = res["Clusters"]
        # choose the correct cluster
        clusters = [c for c in clusters if c["Name"] == cluster_name]
        if clusters:
            break
        elif not "Marker" in res.keys():
            break
        else:
            res = emr.list_clusters(Marker=res["Marker"], ClusterStates=ClusterStates)

    cluster_id = ""
    started = False
    if reuseCluster:
        if not clusters:
            if not startCluster:
                raise Exception("cannot find running EMR cluster: " + cluster_name)
            else:
                (master_ip, cluster_id) = _start_and_wait_for_emr(emr, cluster_name, props, clusterArgs)
                started = True
        else:
            cluster_id = clusters[0]["Id"]
            # possibly the cluster exists but is in the middle of resizing
            _wait_for_cluster_groups(emr, cluster_id)
            master_ip = _get_cluster_ip(emr, cluster_id)

    else:
        if not startCluster:
            raise Exception("If reuseCluster is False , then startCluster must be True")
        else:
            (master_ip, cluster_id) = _start_and_wait_for_emr(emr, cluster_name, props, clusterArgs)
            started = True

    if waitWhenResizing:
        _wait_for_cluster_groups(emr, cluster_id)

    # conn_template = "-s spark -c spark -l python -u http://{}:8998 -t None ADD"
    conn_template = "http://{}:8998"
    conn = conn_template.format(master_ip)

    return (conn, cluster_id, started)


def getSparkSessionInfo(livyUrl: str, appID: str) -> Dict[str, Any]:
    """
    Returns spark session info for each session created (e.g. id, name, owner, appInfo, log, etc.)

    Parameters
    ----------
    livyUrl : str
        The livy url to connect to the spark session.
    appID : str
        The application id of this session.

    Returns
    -------
    sessionInfo : Dict[str, Any]
        JSON formatted data about the spark session.

    Example
    -------
    >>> import aws.utils.notebooks.spark.emr as sparkConnection
    >>> (livy_url, cluster_id, started) = sparkConnection.connect_to_spark(clusterName = "test1",
    ...                                                                    reuseCluster=True,
    ...                                                                    startCluster=True,
    ...                                                                    clusterArgs={})
    >>> appID = spark.sparkContext.applicationId
    >>> getSparkSessionInfo(livyUrl = livy_url, appID = appID)
    """
    sessionInfo = livyUrl + "/sessions"
    r_params = {}
    r = requests.get(url=sessionInfo, params=r_params).json()
    print("session count:" + str(r["total"]))
    sessions = [s for s in r["sessions"] if s["appId"] == appID]
    if len(sessions) > 0:
        return next(sessions)
    else:
        return None


def _get_cluster_id(emr: boto3.client("emr"), clusterName: str) -> str:
    """
    Returns the id of a running cluster with given cluster name.
    """
    clusters = emr.list_clusters()["Clusters"]

    # choose the correct cluster
    clusters = [c for c in clusters if c["Name"] == clusterName and c["Status"]["State"] in ["WAITING", "RUNNING"]]
    if not clusters:
        logger.info("No valid clusters")
        raise Exception("cannot find running cluster: " + clusterName)
    # take the first relevant cluster
    return clusters[0]["Id"]


def _get_cluster_ip(emr: boto3.client("emr"), cluster_id: str, wait_ready: Optional[bool] = True) -> str:
    """
    Waits for instances to be running and then returns the private ip address from the cluster master node.
    """
    if wait_ready:
        _wait_for_cluster_groups(emr, cluster_id)

    response = emr.list_instances(
        ClusterId=cluster_id,
        InstanceGroupTypes=[
            "MASTER",
        ],
    )
    for instance in response["Instances"]:
        if instance["Status"]["State"] == "RUNNING":
            return instance["PrivateIpAddress"]

    if not wait_ready:
        return "not ready"

    raise Exception("Master node is not available: " + str(response["Instances"]))


def get_emr_step_function_error(execution_arn: str) -> None:
    """
    Looks at execution history and gets tries to output error message for EMR cluster's failed start (automatically
    called if EMR fails to start).

    Parameters
    ----------
    execution_arn : str
        The Amazon Resource Name (ARN) of the execution.

    Returns
    -------
    None
        None.
    """
    sfn = boto3.client("stepfunctions")
    logger.error(f"Step function failed launching EMR, execution_arn: {execution_arn}")
    response = sfn.get_execution_history(
        executionArn=execution_arn,
    )
    try:
        state = json.loads(response["events"][len(response["events"]) - 2]["stateEnteredEventDetails"]["input"])
        error = json.loads(state["Error"]["Cause"])
        logger.error(error["errorMessage"])
        raise Exception(error["errorMessage"])
    except:
        logger.error(json.dumps(response))
        raise Exception("Cluster failed starting")


def _start_and_wait_for_emr(
    emr: boto3.client("emr"),
    cluster_name: str,
    props: Dict[str, str],
    clusterArgs: Optional[Dict[str, str]] = dict(),
    wait_ready: Optional[bool] = True,
) -> Tuple[str, str]:
    """
    Start EMR and wait for cluster to begin running as well as get any execution errors.
    """
    sfn = boto3.client("stepfunctions")
    logger.info(f"entering _start_and_wait_for_emr() {cluster_name}")

    emr_functions = _get_emr_functions()
    if len(emr_functions) == 0:
        raise Exception("cannot find any EMR launch functions for this team space")

    if "emr_start_function" in clusterArgs.keys():
        emr_function = emr_functions[clusterArgs["emr_start_function"]]
        del clusterArgs["emr_start_function"]
    else:
        emr_function = list(emr_functions.values())[0]

    clusterArgs["ClusterName"] = cluster_name
    emr_input = {"ClusterConfigOverrides": clusterArgs}
    logger.info("Started EMR with {emr_function} and parameters: %s", emr_input)

    response = sfn.start_execution(stateMachineArn=emr_function["StateMachine"], input=json.dumps(emr_input))
    logger.debug(f"Started EMR step function {response}")

    while True:
        time.sleep(20)
        des_res = sfn.describe_execution(executionArn=response["executionArn"])
        logger.info(f"waiting for EMR to start...")
        if not wait_ready:
            return ("", "")
        if des_res["status"] != "RUNNING":
            break

    logger.info(f"Finished EMR step function")

    if (des_res["status"]) != "SUCCEEDED":
        get_emr_step_function_error(response["executionArn"])

    cluster = json.loads(des_res["output"])
    if "LaunchClusterResult" not in cluster:
        get_emr_step_function_error(response["executionArn"])

    cluster_id = cluster["LaunchClusterResult"]["ClusterId"]

    master_ip = _get_cluster_ip(emr, cluster_id, wait_ready)

    logger.info("cluster ready: %s  ip: %s ", cluster_id, master_ip)

    return (master_ip, cluster_id)


def _get_functions(namespace: Optional[str] = "default", next_token: Optional[str] = None) -> Mapping[str, Any]:
    """
    Returns an EMR Launch Function with its parameters and a next token if more functions exist for a given namespace.
    """
    params = {"Path": f"{SSM_PARAMETER_PREFIX}/{namespace}/"}
    if next_token:
        params["NextToken"] = next_token
    result = boto3.client("ssm").get_parameters_by_path(**params)

    functions = {"EMRLaunchFunctions": [json.loads(p["Value"]) for p in result["Parameters"]]}
    if "NextToken" in result:
        functions["NextToken"] = result["NextToken"]
    return functions


def get_emr_functions() -> List[str]:
    """
    Gets a list of EMR launch functions.

    Parameters
    ----------
    None
        None.

    Returns
    -------
    functions : list
        A list of EMR Launch functions

    Example
    -------
    >>> import aws.utils.notebooks.spark.emr as sparkConnection
    >>> funcs = sparkConnection.get_emr_functions()
    >>> funcs = list(funcs)
    >>> funcs
    """
    return list(_get_emr_functions().keys())


def describe_emr_function(funcName: str) -> Dict[str, Any]:
    """
    Get the EMR Launch Function parameters and values for a given function name.

    Parameters
    ----------
    funcName : str
        Name of the launch function you want information from.

    Returns
    -------
    emr_func : dict
        key-value pairs of different function parameter values.

    Example
    -------
    >>> %%local
    >>> import aws.utils.notebooks.spark.emr as sparkConnection
    >>> emr_func_name = 'standard-default'
    >>> emr_func = sparkConnection.describe_emr_function(emr_func_name)
    >>> emr_func['AllowedClusterConfigOverrides']
    """
    return _get_emr_functions()[funcName]


def _get_emr_functions() -> List[str]:
    """
    Returns a list of all EMR Launch functions and their current values.
    """
    props = get_properties()
    env_name = props["AWS_ORBIT_ENV"]
    team_space = props["AWS_ORBIT_TEAM_SPACE"]
    namespace = f"orbit-{env_name}-{team_space}"
    res = _get_functions(namespace)
    functionList = _get_functions(namespace=namespace)["EMRLaunchFunctions"]
    while "NextToken" in res:
        temp = _get_functions(res["NextToken"])["EMRLaunchFunctions"]
        functionList.extend(temp)

    func = dict()
    for f in functionList:
        func[f["LaunchFunctionName"]] = f

    return func


def _wait_for_cluster_groups(emr: boto3.client("emr"), cluster_id: str) -> None:
    """
    Waits for instance groups in cluster to start running.
    """
    attempts = 30
    while attempts > 0:
        attempts -= 1
        response = emr.list_instance_groups(ClusterId=cluster_id)
        wait = False
        for g in response["InstanceGroups"]:
            if g["Status"]["State"] in (
                "PROVISIONING",
                "BOOTSTRAPPING",
                "RECONFIGURING",
                "RESIZING",
            ):
                logger.info("waiting for cluster group: %s(%s)", g["Name"], g["Status"]["State"])
                wait = True
        if wait:
            time.sleep(30)
        else:
            break


def get_cluster_info(cluster_id: str) -> Dict[str, Dict[str, str]]:
    """
    Gets info on the Master, Core, and Task Nodes for an EMR Cluster.

    Parameters
    ----------
    cluster_id : str
        The EMR cluster ID used in the session

    Returns
    -------
    info : dict
        A dictionary object with information on the master, core, and task nodes existing in a specified cluster.

    Example
    -------
    >>> import aws.utils.notebooks.spark.emr as sparkConnection
    >>> get_cluster_info(cluster_id)
    """
    emr = boto3.client("emr")
    info = {}
    info["MASTER"] = emr.list_instances(ClusterId=cluster_id, InstanceGroupTypes=["MASTER"])
    info["CORE"] = emr.list_instances(ClusterId=cluster_id, InstanceGroupTypes=["CORE"])
    info["TASK"] = emr.list_instances(ClusterId=cluster_id, InstanceGroupTypes=["TASK"])

    return info


def spark_submit(job: Dict[str, Any]) -> Dict[str, Any]:
    """
    This code will run your PySpark job on the spark EMR using your current source directory.

    Parameters
    ----------
    job : dict
        A pyspark job definition to execute

        cluster_id : str
            The EMR cluster ID
        app_name : str
            The Spark application Name
        module : str
            A relative path from repository root to the Python module containing the pyspark application
        waitAppCompletion : bool
            If True , will wait until EMR step is completed.
        app_args : list
            List of arguments to pass to the spark application
        spark_args : list
            List of arguments controlling the spark execution

    Returns
    -------
    response : dict
        The output for the AddJobFlowSteps operation and identifiers of the list of steps added to the job flow.

    Example
    -------
    >>> import aws.utils.notebooks.spark.emr as sparkConnection
    >>> response = sparkConnection.spark_submit(job = {
    ...                                         "cluster_id" : cluster_id,
    ...                                            "app_name":  "test1",
    ...                                            "module" : "samples/python/pyspark/createTbl.py",
    ...                                            "wait_app_completion": False,
    ...                                            "app_args": [arg1,arg2],
    ...                                            "spark_args": [--num-executors,2,--num_cores,4,--executor_memory,1g]
    ...                                             })
    """
    cluster_id = job["cluster_id"] if "cluster_id" in job.keys() else None
    if cluster_id == None:
        raise Exception("cluster_id must be provided")
    app_name = job["app_name"] if "app_name" in job.keys() else None
    if app_name == None:
        raise Exception("app_name must be provided")
    module = job["module"] if "module" in job.keys() else None
    if module == None:
        raise Exception("module must be provided")

    waitAppCompletion = job["wait_app_completion"] if "wait_app_completion" in job.keys() else False
    appargs = job["app_args"] if "app_args" in job.keys() else []
    sparkargs = job["spark_args"] if "spark_args" in job.keys() else []
    props = get_properties()
    if waitAppCompletion:
        waitApp = "true"
    else:
        waitApp = "false"

    workspaceDir = "workspace"

    notebookInstanceName = socket.gethostname()

    s3WorkspaceDir = "s3://{}/{}/workspaces/{}".format(
        props["AWS_ORBIT_S3_BUCKET"],
        props["AWS_ORBIT_TEAM_SPACE"],
        notebookInstanceName,
    )
    cmd = 'aws s3 sync --delete --exclude "*.git/*" {} {}'.format(workspaceDir, s3WorkspaceDir)
    logout = os.popen(cmd).read()
    logger.info("s3 workspace directory is %s", s3WorkspaceDir)
    logger.debug(logout)

    module = os.path.join(s3WorkspaceDir, module)

    emr = boto3.client("emr")
    args = [
        "/usr/bin/spark-submit",
        "--verbose",
        "--deploy-mode",
        "cluster",
        "--master",
        "yarn",
        "--conf",
        "pyspark.yarn.submit.waitAppCompletion=" + waitApp,
        "--conf",
        "hive.metastore.client.factory.class=com.amazonaws.glue.catalog.metastore.AWSGlueDataCatalogHiveClientFactory",
        "--conf",
        "hive.metastore.connect.retries=5",
    ]
    args.extend(sparkargs)
    args.append(module)
    args.extend(appargs)

    response = emr.add_job_flow_steps(
        JobFlowId=cluster_id,
        Steps=[
            {
                "Name": app_name,
                "ActionOnFailure": "CONTINUE",
                "HadoopJarStep": {"Jar": "command-runner.jar", "Args": args},
            },
        ],
    )

    return response


def get_team_clusters(cluster_id: Optional[str] = None) -> Dict[str, Dict[str, str]]:
    """
    Finds all running clusters for a user' team space and returns cluster-level details including status, hardware
    and software configuration, VPC settings, and so on for each cluster.

    Parameters
    ----------
    cluster_id : str, optional
        The unique EMR cluster ID to get information about.

    Returns
    -------
    clusters_info : dict
        A dictionary of one or many clusters and their configuration.

    Example
    -------
    >>> import aws.utils.notebooks.spark.emr as sparkConnection
    >>> sparkConnection.get_team_clusters()
    """

    emr = boto3.client("emr")
    props = get_properties()
    if cluster_id == None:
        clusters = emr.list_clusters(ClusterStates=["STARTING", "BOOTSTRAPPING", "RUNNING", "WAITING"])
        if "Clusters" not in clusters:
            raise Exception("Error calling list_clusters()")
        if len(clusters["Clusters"]) == 0:
            logger.info("no emr clusters found for team space")
            return {}
        clusters = clusters["Clusters"]
    else:
        clusters = [{"Id": cluster_id}]

    clusters_info = {}
    for cluster in clusters:
        cluster_id = cluster["Id"]
        cluster_info = emr.describe_cluster(ClusterId=cluster_id)
        if "Cluster" not in cluster_info:
            raise Exception("Error calling describe_cluster()")

        tag_list = cluster_info["Cluster"]["Tags"]
        tags = {}
        for tag in tag_list:
            tags[tag["Key"]] = tag["Value"]

        if ORBIT_PRODUCT_KEY not in tags or tags[ORBIT_PRODUCT_KEY] != ORBIT_PRODUCT_NAME:
            continue

        if tags[ORBIT_ENV] != props["AWS_ORBIT_ENV"] or tags[AWS_ORBIT_TEAM_SPACE] != props["AWS_ORBIT_TEAM_SPACE"]:
            continue

        cluster_nodes_info = get_cluster_info(cluster_id)
        ip = _get_cluster_ip(emr, cluster_id, False)
        livy_url = f"http://{ip}:8998"
        cluster_model = {}
        cluster_model["cluster_id"] = cluster_id
        cluster_model["livy_url"] = livy_url
        cluster_model["ip"] = ip
        cluster_model["Name"] = cluster["Name"]
        cluster_model["State"] = cluster["Status"]["State"]
        cluster_model["info"] = cluster_info
        cluster_model["dashboard_link"] = "http://tbd"
        cluster_model["instances"] = cluster_nodes_info
        clusters_info[cluster_id] = cluster_model

    return clusters_info
